Fast Log-Domain Sinkhorn Optimal Transport
with Warp-Level GPU Reductions
Abstract
Entropic regularized optimal transport (OT) via the Sinkhorn algorithm has become a fundamental tool in machine learning, yet existing implementations either suffer from numerical instability for small regularization parameters or incur significant overhead from deep learning frameworks. We present FastSinkhorn, a lightweight, native CUDA implementation of the log-domain Sinkhorn algorithm that combines warp-level shuffle reductions with shared-memory tiling to achieve high GPU utilization without sacrificing numerical stability. Our solver operates entirely in the log-domain, enabling robust computation for regularization parameters as small as where standard-domain methods fail. On dense OT problems with , our implementation achieves speedup over the widely-used POT library and speedup over GPU-accelerated PyTorch baselines, while consuming only 256 MB of GPU memory. We validate our solver on image color transfer, 3D point cloud matching, and convergence analysis, demonstrating that native CUDA kernels with careful numerical treatment provide a practical and efficient foundation for large-scale optimal transport computation.
1 Introduction
Optimal transport (OT) provides a mathematically principled framework for comparing probability distributions by computing the minimum-cost coupling between them (Villani, 2003, 2009). In the discrete setting, the Kantorovich formulation reduces to a linear program over coupling matrices, with applications spanning generative modeling (Arjovsky et al., 2017), domain adaptation (Courty et al., 2017), document similarity (Kusner et al., 2015), and geometric processing (Solomon et al., 2015).
The computational bottleneck of exact OT solvers—scaling as via the network simplex algorithm—was dramatically alleviated by Cuturi (2013), who introduced entropic regularization and showed that the resulting problem can be solved via Sinkhorn’s matrix scaling algorithm in near-linear time per iteration. This breakthrough has led to widespread adoption of the Sinkhorn algorithm in machine learning pipelines (Peyré and Cuturi, 2019).
Despite its popularity, practical deployment of the Sinkhorn algorithm on modern hardware faces two key challenges:
Numerical instability.
The standard-domain Sinkhorn algorithm computes the Gibbs kernel , which causes catastrophic overflow or underflow in floating-point arithmetic when the regularization parameter is small—precisely the regime needed for accurate OT approximation. Log-domain stabilization (Schmitzer, 2019) addresses this but has not been widely adopted in GPU-optimized implementations.
Framework overhead.
Existing GPU implementations rely on deep learning frameworks (PyTorch, JAX) or specialized libraries (KeOps/GeomLoss (Feydy et al., 2019; Charlier et al., 2021)), which introduce significant overhead from automatic differentiation graphs, memory allocators, and Python-to-GPU dispatch. For applications that require only the forward OT computation—such as point cloud registration, color transfer, or batch distance computation—this overhead is unnecessary.
1.1 Contributions
We present FastSinkhorn, a lightweight CUDA C++ library for entropic regularized optimal transport. Our contributions are:
-
1.
Numerically stable GPU solver. A log-domain Sinkhorn implementation with a two-pass LogSumExp reduction that operates entirely in log-space, handling regularization parameters as small as without numerical failure.
-
2.
Warp-level GPU optimizations. We exploit CUDA warp shuffle instructions (__shfl_down_sync) for intra-warp reductions, combined with shared memory for cross-warp communication. This hierarchical reduction eliminates shared-memory bank conflicts within warps and reduces synchronization overhead by compared to pure shared-memory reductions.
-
3.
Comprehensive evaluation. We provide extensive benchmarks against POT (Flamary et al., 2021), GeomLoss (Feydy et al., 2019), and PyTorch-based solvers, along with ablation studies quantifying the contribution of each optimization, numerical stability analysis, and real-world applications in image color transfer and point cloud matching.
Our solver is self-contained (no dependencies beyond the CUDA runtime), supports both pre-computed cost matrices and Euclidean point cloud inputs, and is publicly available under the MIT license.
2 Related Work
Entropic optimal transport.
The use of entropic regularization for optimal transport was popularized by Cuturi (2013), who showed that adding a KL-divergence penalty transforms the linear program into a strictly convex problem solvable via Sinkhorn iterations (Sinkhorn, 1967). The comprehensive treatment by Peyré and Cuturi (2019) establishes the theoretical foundations, including convergence rates and approximation guarantees. Altschuler et al. (2017) provided near-linear time complexity bounds, showing that operations suffice for an -approximate solution.
Numerical stabilization.
The standard Sinkhorn algorithm suffers from numerical overflow/underflow when is small, as the Gibbs kernel entries span extreme ranges. Schmitzer (2019) proposed log-domain stabilization, and Chizat et al. (2018) extended this to unbalanced settings. While these stabilizations are well-understood theoretically, many GPU implementations still default to the numerically fragile standard-domain formulation.
GPU-accelerated solvers.
Several GPU-based OT solvers exist in the literature. The POT library (Flamary et al., 2021) provides a comprehensive Python API with optional GPU acceleration via CuPy. GeomLoss (Feydy et al., 2019) and its KeOps backend (Charlier et al., 2021) offer GPU-accelerated Sinkhorn divergences with automatic differentiation, designed for integration with PyTorch. These framework-based approaches provide flexibility but carry overhead from the Python runtime, autograd engine, and generic GPU memory management.
Algorithmic acceleration.
Beyond GPU parallelization, several algorithmic improvements reduce the computational cost of OT. Multiscale approaches (Gerber and Maggioni, 2017; Schmitzer, 2019) use coarse-to-fine strategies with sparse representations. Low-rank factorizations (Scetbon et al., 2021) approximate the transport plan with reduced-rank factors, achieving sub-quadratic complexity. Accelerated gradient methods (Dvurechensky et al., 2018; Lin et al., 2019) provide improved convergence rates for the dual problem. These approaches are complementary to our GPU optimization strategy and could be combined in future work.
Our position.
Our work occupies a distinct niche: a native CUDA implementation that combines log-domain stability with warp-level GPU optimizations, targeting the common case of dense, moderate-sized () OT problems where the full cost matrix fits in GPU memory. Unlike framework-based approaches, our solver has zero Python overhead and minimal memory footprint, making it suitable for latency-sensitive applications and embedding in C++ pipelines.
3 Background
3.1 Discrete Optimal Transport
Let and be discrete probability distributions on point sets and , where denotes the probability simplex. Given a cost matrix with , the Kantorovich optimal transport problem seeks the minimum-cost coupling:
| (1) |
where the feasible set of couplings (transport plans) is:
| (2) |
This is a linear program in variables with equality constraints. Exact solvers based on the network simplex algorithm have complexity, which is prohibitive for large (Peyré and Cuturi, 2019).
3.2 Entropic Regularization
Following Cuturi (2013), we consider the entropically regularized problem:
| (3) |
where is the regularization parameter and the KL divergence is:
| (4) |
The addition of the strictly convex KL term ensures a unique minimizer . As , we recover the solution to the original problem: (Peyré and Cuturi, 2019).
3.3 Structure of the Optimal Solution
Writing the KKT conditions with dual variables (for row constraints) and (for column constraints), and setting , the optimal transport plan has the Gibbs kernel structure:
| (5) |
3.4 The Sinkhorn Algorithm
Substituting Eq. (5) into the marginal constraints yields a fixed-point iteration. The Sinkhorn algorithm alternates:
| (6) | ||||
| (7) |
starting from , .
Theorem 1 (Linear convergence (Peyré and Cuturi, 2019, Theorem 4.2)).
The Sinkhorn iterates converge linearly: , where and is the cost range.
The contraction rate approaches 1 as , revealing the fundamental tension: smaller yields a better OT approximation but requires more iterations.
4 Method: GPU-Optimized Log-Domain Sinkhorn
4.1 Log-Domain Stabilization
The standard Sinkhorn algorithm (Section 3) computes the Gibbs kernel explicitly. For typical cost matrices with and , the kernel entries span —far beyond the range of IEEE 754 single-precision floats (). This causes catastrophic underflow.
We instead work entirely in the log-domain. The key operation is the numerically stable LogSumExp:
| (8) |
By subtracting the maximum before exponentiation, the largest exponential becomes (no overflow) and all others satisfy (no overflow). The sum is at least 1, so the logarithm is non-negative (no underflow issue).
4.2 CUDA Parallelization Strategy
Each dual update requires computing one LogSumExp per row (for ) or per column (for ). We assign one CUDA thread block per row/column, with threads within the block cooperating on the reduction. This provides (or ) independent blocks—sufficient to saturate modern GPUs for .
Algorithm 1 presents the complete procedure.
4.3 Warp-Level Reductions
The LogSumExp computation requires two reductions per row/column: a max-reduction (Pass 1) and a sum-reduction (Pass 2). We implement these using a hierarchical reduction scheme:
Intra-warp reduction.
Within each warp (32 threads), we use the __shfl_down_sync instruction to perform butterfly reductions in steps, without any shared memory access or synchronization barriers. For the max-reduction:
| (11) |
for offsets . The sum-reduction replaces with .
Cross-warp reduction.
With block size (8 warps), each warp’s result is written to shared memory (one slot per warp). The first warp then performs a final warp-level reduction across these 8 values.
This two-level design eliminates shared memory bank conflicts for intra-warp communication, avoids __syncthreads() barriers within warps, and requires only shared memory slots regardless of block size.
4.4 Memory Layout and Access Patterns
The cost matrix dominates memory usage at floats, while dual potentials require only —negligible in comparison. For , the cost matrix occupies 256 MB and the potentials add 64 KB.
Coalesced access.
The cost matrix is stored in row-major order. In the -update kernel, block reads row of sequentially—threads within a warp access consecutive addresses, achieving coalesced global memory loads. The -update kernel accesses column with stride , which is less favorable; however, the L2 cache (typically 4–6 MB on modern GPUs) mitigates this when the cost matrix fits in cache.
Precomputed log-weights.
We precompute and on the host and transfer them once, avoiding repeated logarithm computations on the GPU.
4.5 Convergence Monitoring
The marginal error , where is the row marginal, requires a GPU-to-CPU data transfer for the scalar result. To amortize this synchronization cost, we check convergence every iterations (configurable). The row marginal is computed in log-domain:
| (12) |
which reuses the same LogSumExp kernel structure as the dual updates.
4.6 Complexity Analysis
Each Sinkhorn iteration performs two LogSumExp reductions over the cost matrix, giving work per iteration. With iterations and convergence checks every iterations, the total complexity is . Since is a small constant, this simplifies to . On the GPU, the (or ) independent reductions execute in parallel across blocks, achieving an effective wall-clock complexity of (or ) per SM, distributed across the available streaming multiprocessors.
5 Experiments
We evaluate FastSinkhorn across six experimental dimensions: baseline comparison, scaling behavior, ablation study, numerical stability, convergence analysis, and real-world applications. Our CUDA solver is benchmarked on an NVIDIA RTX 3090 (24 GB memory, 82 streaming multiprocessors, compute capability 8.6). Python baselines (POT, GeomLoss, PyTorch Sinkhorn) are benchmarked on an NVIDIA Tesla T4 (16 GB).
5.1 Baseline Comparisons
We compare against three widely-used implementations:
All methods solve the same problem: OT between two random distributions on a uniform 1D grid with squared Euclidean cost and . We measure wall-clock time (excluding data transfer) averaged over 10 runs after 3 warmup runs.
| Method | ||||||
|---|---|---|---|---|---|---|
| POT (CPU) | 3.2 | 13.1 | 55.1 | 256.0 | 1097.2 | 4462.1 |
| GeomLoss (T4) | 37.1 | 24.1 | 23.5 | 80.0 | 304.8 | 1193.5 |
| PyTorch Sinkhorn (T4) | 10.5 | 10.3 | 36.6 | 135.9 | 533.5 | 2174.5 |
| Ours (RTX 3090) | 0.3 | 1.3 | 5.1 | 20.7 | 89.1 | 371.6 |
| Speedup vs. POT | 10.0 | 10.4 | 10.8 | 12.4 | 12.3 | 12.0 |
| Speedup vs. PyTorch | 32.4 | 8.2 | 7.2 | 6.6 | 6.0 | 5.9 |
Table 1 and Figure 1 show the results. Our native CUDA implementation consistently outperforms all baselines. Compared to POT (CPU), we achieve – speedup, with the gap widening at larger problem sizes. Compared to PyTorch Sinkhorn on a Tesla T4, we achieve – speedup, demonstrating the benefit of eliminating framework overhead. Note that our solver runs on an RTX 3090 while GPU baselines run on a T4; even accounting for this hardware difference (the RTX 3090 has approximately higher FP32 throughput), our approach retains a significant advantage.
5.2 Scaling Behavior
We evaluate our solver for problem sizes from to with . Figure 2 shows the results. Computation time scales quadratically with , consistent with the per-iteration cost of dense matrix operations. Peak GPU memory is dominated by the cost matrix (e.g., 1024 MB for ). The iteration count remains approximately constant (180–200) across problem sizes, confirming that the convergence rate of Sinkhorn’s algorithm depends on rather than (Theorem 1). Our solver handles in 1.54 seconds on a single RTX 3090.
5.3 Ablation Study
We ablate each optimization to quantify its individual contribution, using and as the reference configuration.
| Configuration | Time (ms) | Slowdown |
|---|---|---|
| Full system (ours) | 14.8 | |
| Shared-memory only (no warp shuffle) | 28.5 | |
| Standard-domain (no log-LSE)† | 16.9 | |
| Block size = 64 (vs. 256) | 21.1 | |
| Block size = 128 | 16.7 | |
| Block size = 512 | 16.8 | |
| Convergence check every iteration | 24.8 | |
| Check interval = 5 | 16.8 | |
| Check interval = 10 | 15.2 |
† Standard-domain solver fails (NaN) for ; time reported for only.
Table 2 reveals that warp-level shuffle reductions provide the single largest speedup (), followed by the convergence check interval optimization ( when checking every iteration vs. every 20 iterations). The optimal block size is 256, with smaller blocks (64) suffering from insufficient parallelism and larger blocks (512) incurring register pressure. The log-domain formulation adds only overhead compared to the standard-domain variant, a modest price for the substantially improved numerical stability (Section 5.4).
5.4 Numerical Stability
A critical advantage of our log-domain formulation is robustness for small . We sweep from to with and compare:
- •
-
•
Standard-domain: computes Gibbs kernel and scaling vectors .
Figure 4 shows the results. The standard-domain solver produces NaN for due to overflow in the Gibbs kernel . In contrast, our log-domain solver remains numerically stable and converges successfully for all tested values down to . Both methods produce comparable transport costs for , confirming that the log-domain formulation does not sacrifice accuracy.
5.5 Convergence Analysis
We record the marginal error at each convergence check to visualize the convergence profile across different problem sizes and regularization parameters.
5.6 Applications
We demonstrate practical utility on two tasks:
Image color transfer.
Given a source and target image, we sample pixels from each, compute the OT plan between their RGB distributions (), and apply barycentric mapping to transfer the color palette. Figure 6 shows qualitative results: the warm color palette of the source image is successfully transferred to match the cool target palette.
3D point cloud matching.
We compute OT correspondences between two 3D point clouds ( points each), where the target is a rotated and translated copy of the source with added Gaussian noise. Figure 7 shows that the OT plan recovers accurate correspondences between the two clouds.
6 Conclusion
We presented FastSinkhorn, a native CUDA implementation of the log-domain Sinkhorn algorithm for entropic regularized optimal transport. By combining numerically stable LogSumExp computation with warp-level shuffle reductions and shared-memory tiling, our solver achieves speedup over the POT library and over PyTorch-based GPU solvers on dense OT problems up to . The log-domain formulation enables robust computation for regularization parameters as small as , where standard-domain methods fail due to floating-point overflow.
Limitations.
Our implementation has several limitations. First, it operates exclusively in single precision (float32), which may be insufficient for problems requiring very high accuracy. Second, the dense cost matrix storage limits scalability to approximately on a 16 GB GPU. Third, our solver does not support automatic differentiation, precluding direct use in gradient-based learning pipelines.
Future work.
Several extensions are promising: (1) mixed-precision (fp16/bf16) computation for the Gibbs kernel, with fp32 accumulation for the LogSumExp; (2) sparse or low-rank cost matrix representations to scale beyond ; (3) a pybind11 wrapper for seamless Python integration; (4) integration with multiscale strategies (Schmitzer, 2019) for further acceleration.
Reproducibility.
All code, experiment scripts, and plotting utilities are publicly available at https://github.com/xiao98/Fast-Sinkhorn-CUDA under the MIT license.
References
- Near-linear time approximation algorithms for optimal transport via Sinkhorn iteration. In Advances in Neural Information Processing Systems (NeurIPS), Vol. 30. Cited by: §2.
- Wasserstein generative adversarial networks. In Proceedings of the 34th International Conference on Machine Learning (ICML), pp. 214–223. Cited by: §1.
- Kernel operations on the GPU, with autodiff, without memory overflows. In Journal of Machine Learning Research, Vol. 22, pp. 1–6. Cited by: §1, §2.
- Scaling algorithms for unbalanced optimal transport problems. Mathematics of Computation 87 (314), pp. 2563–2609. Cited by: §2.
- Optimal transport for domain adaptation. In IEEE Transactions on Pattern Analysis and Machine Intelligence, Vol. 39, pp. 1853–1865. Cited by: §1.
- Sinkhorn distances: lightspeed computation of optimal transport. In Advances in Neural Information Processing Systems (NeurIPS), Vol. 26. Cited by: §1, §2, §3.2.
- Computational optimal transport: complexity by accelerated gradient descent is better than by Sinkhorn’s algorithm. In Proceedings of the 35th International Conference on Machine Learning (ICML), pp. 1367–1376. Cited by: §2.
- Interpolating between optimal transport and MMD using Sinkhorn divergences. In Proceedings of the 22nd International Conference on Artificial Intelligence and Statistics (AISTATS), pp. 2681–2690. Cited by: item 3, §1, §2, 2nd item.
- POT: Python optimal transport. Journal of Machine Learning Research 22 (78), pp. 1–8. Cited by: item 3, §2, 1st item.
- Scaling of matrices to achieve specified row and column sums. Numer. Math. 53, pp. 139–148. Cited by: §A.2.
- Multiscale strategies for computing optimal transport. In Journal of Machine Learning Research, Vol. 18, pp. 1–32. Cited by: §2.
- From word embeddings to document distances. In Proceedings of the 32nd International Conference on Machine Learning (ICML), pp. 957–966. Cited by: §1.
- On efficient optimal transport: an analysis of greedy and accelerated mirror descent algorithms. In Proceedings of the 36th International Conference on Machine Learning (ICML), pp. 3982–3991. Cited by: §2.
- Computational optimal transport: with applications to data science. Foundations and Trends in Machine Learning 11 (5–6), pp. 355–607. Cited by: §1, §2, §3.1, §3.2, Theorem 1.
- Low-rank Sinkhorn factorization. In Proceedings of the 38th International Conference on Machine Learning (ICML), pp. 9344–9354. Cited by: §2.
- Stabilized sparse scaling algorithms for entropy regularized transport problems. SIAM Journal on Scientific Computing 41 (3), pp. A1443–A1481. Cited by: §1, §2, §2, §6.
- Diagonal equivalence to matrices with prescribed row and column sums. The American Mathematical Monthly 74 (4), pp. 402–405. Cited by: §2.
- Convolutional Wasserstein distances: efficient optimal transport on geometric domains. ACM Transactions on Graphics (SIGGRAPH) 34 (4), pp. 66:1–66:11. Cited by: §1.
- Topics in optimal transportation. Graduate Studies in Mathematics, Vol. 58, American Mathematical Society. Cited by: §1.
- Optimal transport: old and new. Grundlehren der mathematischen Wissenschaften, Vol. 338, Springer. Cited by: §1.
Appendix A Mathematical Derivation
We provide a complete derivation from the Kantorovich problem to the GPU-optimized log-domain Sinkhorn algorithm.
A.1 From KKT Conditions to Sinkhorn Updates
The Lagrangian of the regularized problem (3) is:
| (13) |
Setting :
| (14) |
Substituting into the row marginal constraint :
| (15) | ||||
| (16) | ||||
| (17) |
The column constraint yields the analogous update for .
A.2 Convergence Rate
Proof sketch of Theorem 1.
Define the operator by the Sinkhorn updates. The operator is a contraction in the Hilbert projective metric on the positive cone. The contraction rate is determined by the Birkhoff contraction coefficient:
| (18) |
where . This follows from classical results on positive matrix scaling [Franklin and Lorenz, 1989]. ∎
A.3 LogSumExp Numerical Analysis
The standard computation fails when (overflow of in float32) or when and all terms underflow to zero (yielding ). In our setting, , and for with , we have —clearly problematic.
The stabilized version with ensures:
-
•
The largest exponent is (no overflow).
-
•
All exponents satisfy (no overflow).
-
•
The sum is at least 1 (from the achieving the max), so (stable).
Appendix B Extended Benchmark Results
| Time (ms) | Iterations | Transport cost | Converged | ||
|---|---|---|---|---|---|
| 256 | 0.1 | 0.15 | 31 | 0.0413 | ✓ |
| 256 | 0.01 | 0.32 | 97 | 0.0395 | ✓ |
| 256 | 0.001 | 0.65 | 340 | 0.0391 | ✓ |
| 512 | 0.1 | 0.64 | 30 | 0.0391 | ✓ |
| 512 | 0.01 | 1.26 | 80 | 0.0366 | ✓ |
| 512 | 0.001 | 2.41 | 284 | 0.0406 | ✓ |
| 1024 | 0.1 | 2.46 | 27 | 0.0429 | ✓ |
| 1024 | 0.01 | 5.09 | 100 | 0.0372 | ✓ |
| 1024 | 0.001 | 9.99 | 319 | 0.0377 | ✓ |
| 2048 | 0.1 | 10.88 | 29 | 0.0394 | ✓ |
| 2048 | 0.01 | 20.66 | 118 | 0.0400 | ✓ |
| 2048 | 0.001 | 40.26 | 342 | 0.0376 | ✓ |
| 4096 | 0.1 | 44.68 | 25 | 0.0373 | ✓ |
| 4096 | 0.01 | 89.09 | 107 | 0.0403 | ✓ |
| 4096 | 0.001 | 175.00 | 306 | 0.0370 | ✓ |
| 8192 | 0.1 | 176.51 | 30 | 0.0421 | ✓ |
| 8192 | 0.01 | 371.60 | 82 | 0.0406 | ✓ |
| 8192 | 0.001 | 714.88 | 294 | 0.0412 | ✓ |
Appendix C GPU Specifications
| Property | Value |
|---|---|
| Device | NVIDIA GeForce RTX 3090 |
| Compute capability | 8.6 |
| Streaming multiprocessors | 82 |
| CUDA cores | 10496 |
| Memory | 24 GB GDDR6X |
| Memory bandwidth | 936 GB/s |
| CUDA version | 12.2 |
| Baseline GPU (Python baselines) | |
| Device | NVIDIA Tesla T4 |
| Compute capability | 7.5 |
| Streaming multiprocessors | 40 |
| CUDA cores | 2560 |
| Memory | 16 GB GDDR6 |
| Memory bandwidth | 320 GB/s |
Appendix D Core Kernel Implementation
For reference, we include the complete CUDA kernel for the -update (the -update is analogous):
__global__ void updateAlphaKernel(
const float* C, const float* log_nu,
const float* beta, float* alpha,
int n, int m, float eps)
{
const int i = blockIdx.x;
if (i >= n) return;
const float inv_eps = 1.0f / eps;
const float* Ci = C + i * m;
// Pass 1: max for numerical stability
float thread_max = -FLT_MAX;
for (int j = threadIdx.x; j < m; j += blockDim.x) {
float val = (beta[j] - Ci[j]) * inv_eps + log_nu[j];
thread_max = fmaxf(thread_max, val);
}
float row_max = blockReduceMax(thread_max);
__shared__ float s_max;
if (threadIdx.x == 0) s_max = row_max;
__syncthreads();
row_max = s_max;
// Pass 2: stable sum
float thread_sum = 0.0f;
for (int j = threadIdx.x; j < m; j += blockDim.x) {
float val = (beta[j] - Ci[j]) * inv_eps + log_nu[j];
thread_sum += expf(val - row_max);
}
float row_sum = blockReduceSum(thread_sum);
if (threadIdx.x == 0) {
alpha[i] = -eps * (row_max
+ logf(fmaxf(row_sum, 1e-30f)));
}
}